//************************************************************************************************************************** 
//* BlueBase64
//* Copyright (c) 2002 Josh Harler
//* 
//* This software is provided 'as-is', without any express or implied warranty. In no event
//* will the authors be held liable for any damages arising from the use of this software.
//* 
//* Permission is granted to anyone to use this software for any purpose, including commercial
//* applications, and to alter it and redistribute it freely, subject to the following restrictions:
//* 
//* 	1. The origin of this software must not be misrepresented; you must not claim that you
//* 	wrote the original software. If you use this software in a product, an acknowledgment in the
//* 	product documentation would be appreciated but is not required.
//* 
//* 	2. Altered source versions must be plainly marked as such, and must not be misrepresented as
//* 	being the original software.
//* 
//* 	3. This notice may not be removed or altered from any source distribution.
//*
//*
//* file   Main.cpp
//**

// Private Headers =========================================================================================================

#include <iostream>

#include "Blue/Data/File.h"
#include "Blue/Data/InputStream.h"
#include "Blue/Data/DiskFileSystem.h"
#include "Blue/Util/CommandLineArgs.h"

#include "Base64EncodeFilter.h"
#include "Base64DecodeFilter.h"
#include "Base64LineFormatter.h"


// Private Defines/Enums/Typedefs/Etc ======================================================================================

using namespace std;

using namespace blue;
using namespace blue::common;
using namespace blue::data;
using namespace blue::util;


// Private Classes/Structs =================================================================================================

// Private Global Variables ================================================================================================

namespace {
	const String APP_NAME    = "BlueBase64";
	const String APP_VERSION = "1.0";
}


// External Global Variables ===============================================================================================

// Private Functions =======================================================================================================

ostream& operator<<( ostream& out, const String& str )
{
	out.write(str.getAsCStr(), str.getLength());
	return (out);
}


// Functions ===============================================================================================================

int main( int argc, char** argv )
{
	int retVal = 0;

	try {

		// setup command line arguments
		CommandLineArgs cmdLine(argc, argv);
		cmdLine.setUsageLine("Base64.exe [-d|-e] <input file> [output file]");
		cmdLine.addOption("Help",    "?", "help",    false, "Displays this screen");
		cmdLine.addOption("Version", "v", "version", false, "Displays the current version");
		cmdLine.addOption("Encode",  "e", "encode",  false, "Encodes the input file");
		cmdLine.addOption("Decode",  "d", "decode",  false, "Decodes the input file (default)");
		cmdLine.processArgs();

		// get settings
		if( argc == 1 || cmdLine.isOptionSet("Help") ) {
			cout << cmdLine.displayUsage() << endl;
			return (argc == 1 ? 1 : 0);
		}

		if( cmdLine.isOptionSet("Version") ) {
			cout << APP_NAME << " version " << APP_VERSION << endl;
			return (0);
		}

		bool decoding = true;

		if( cmdLine.isOptionSet("Encode") ) {
			if( cmdLine.isOptionSet("Decode") ) {
				throw CommandLineArgsInvalidException("Either encode or decode the file, don't do both!");
			}
			else {
				decoding = false;
			}
		}

		Array<String> trailing = cmdLine.getTrailingArgs();

		if( trailing.getSize() < 1 ) {
			throw CommandLineArgsInvalidException("Input file is required for processing");
		}
		if( trailing.getSize() > 2 ) {
			throw CommandLineArgsInvalidException("Only give input file and output file");
		}

		String inputFileName = trailing[0];
		String outputFileName = (trailing.getSize() == 2 ? trailing[1] : inputFileName);


		// get the files
		String actlOutputFileName = outputFileName;

		if( inputFileName == outputFileName ) {
			actlOutputFileName = DiskFileSystem::instance()->getTempDirectory();
			actlOutputFileName += "/" + APP_NAME + APP_VERSION + "." + DiskFileSystem::instance()->fileFormatFileName(outputFileName, FileSystem::FMT_RELATIVE);
		}

		File inputFile(inputFileName, File::READ);
		File outputFile(actlOutputFileName, File::WRITEOVER);

		// create the input stream chain
		InputStream input(&inputFile);

		FilterChain chain;
		if( decoding ) {
			chain.addFilter( new Base64DecodeFilter(), true );
		}
		else {
			chain.addFilter( new Base64EncodeFilter(), true );
			chain.addFilter( new Base64LineFormatter(62), true );
		}

		input.setFilterChain(&chain);
		input.startReading();

		// this try/catch block is to catch invalid base64 data when decoding
		try {
			// filter the input and write it to the output
			while( input.isReading() ) {
				uint8_t byte;
				if( input.read(&byte, sizeof(byte)) != sizeof(byte) ) {
					break;
				}
				outputFile.write(&byte, sizeof(byte));
			}
		}
		catch( Exception& e ) {
			cout << e.getException() << ": " << e.getDescription() << endl;
			retVal = 1;
		}

		input.stopReading();

		inputFile.close();
		outputFile.close();

		if( actlOutputFileName != outputFileName ) {
			// move the written file to the new location
			if( File::exists(outputFileName) ) {
				File(outputFileName).remove();
			}

			outputFile.move(outputFileName);
		}
	}
	catch( FileAccessDeniedException& e ) {
		cout << "Cannot access file: " + e.getFileName() << endl; return (1);
	}
	catch( Exception& e ) {
		cout << e.getException() << ": " << e.getDescription() << endl; return (1);
	}
	catch( ... ) {
		cout << "An unknown exception was thrown" << endl; return (1);
	}

	return (retVal);
}
